@@ -1,5 +1,5 @@
-# sigtools - Python module to manipulate function signatures
-# Copyright (c) 2013 Yann Kaiser
+# sigtools - Collection of Python modules for manipulating function signatures
+# Copyright (C) 2013-2022 Yann Kaiser
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -19,19 +19,29 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
-"""Decorators to modify a callable's signature."""
+"""
+`sigtools.modifiers`: Modify the effective signature of the decorated callable
+------------------------------------------------------------------------------
+
+The functions in this module can be used as decorators to mark and enforce some
+parameters to be keyword-only (`kwoargs`) or annotate (`annotate`) them, just
+like you can :ref:`using Python 3 syntax <def>`. You can also mark and enforce
+parameters to be positional-only (`posoargs`). `autokwoargs` helps you quickly
+make your parameters with default values become keyword-only.
+
+"""
 
 from functools import partial, update_wrapper
 
-import six
+from sigtools import _util, _specifiers, _signatures
 
-from sigtools import _util
-
-__all__ = ['annotate', 'posoargs', 'kwoargs', 'autokwoargs']
+__all__ = ['annotate', 'kwoargs', 'autokwoargs', 'posoargs']
 
 
 class _PokTranslator(_util.OverrideableDataDesc):
-    def __new__(cls, func=None, posoargs=(), kwoargs=(), *args, **kwargs):
+    __slots__ = ['__self__', 'func', 'posoarg_names', 'kwoarg_names', 'kwopos', '__signature__']
+
+    def __new__(cls, func=None, posoargs=(), kwoargs=(), **kwargs):
         if func is None:
             return partial(_PokTranslator, posoargs=posoargs,
                            kwoargs=kwoargs, **kwargs)
@@ -39,11 +49,20 @@ class _PokTranslator(_util.OverrideableDataDesc):
             return super(_PokTranslator, cls).__new__(cls)
         return func
 
-    def __init__(self, func, posoargs=(), kwoargs=(), *args, **kwargs):
+    def __init__(self, func, posoargs=(), kwoargs=(), **kwargs):
+        update_wrapper(self, func)
+        try:
+            self.__self__ = func.__self__
+        except AttributeError:
+            pass
+        try:
+            del self._sigtools__forger
+        except AttributeError:
+            pass
+        super(_PokTranslator, self).__init__(**kwargs)
         self.func = func
         self.posoarg_names = set(posoargs)
         self.kwoarg_names = set(kwoargs)
-        super(_PokTranslator, self).__init__(**kwargs)
         if isinstance(func, _PokTranslator):
             self._merge_other(func)
         self._prepare()
@@ -58,8 +77,6 @@ class _PokTranslator(_util.OverrideableDataDesc):
             self.custom_getter, other.custom_getter)
 
     def _prepare(self):
-        update_wrapper(self, self.func, updated=())
-
         intersection = self.posoarg_names & self.kwoarg_names
         if intersection:
             raise ValueError(
@@ -67,7 +84,7 @@ class _PokTranslator(_util.OverrideableDataDesc):
                 + ' '.join(repr(name) for name in intersection))
         to_use = self.posoarg_names | self.kwoarg_names
 
-        sig = _util.signature(self.func)
+        sig = _specifiers.forged_signature(self.func, auto=False)
         params = []
         kwoparams = []
         kwopos = self.kwopos = []
@@ -93,9 +110,14 @@ class _PokTranslator(_util.OverrideableDataDesc):
                     params.append(param)
             else: # not a POK param
                 if param.name in to_use:
-                    raise ValueError(
-                        '{0.name!r} is not of kind POSITIONAL_OR_KEYWORD, but:'
-                        ' {0.kind}'.format(param))
+                    if param.kind == param.POSITIONAL_ONLY and param.name in self.posoarg_names:
+                        to_use.remove(param.name)
+                    elif param.kind == param.KEYWORD_ONLY and param.name in self.kwoarg_names:
+                        to_use.remove(param.name)
+                    else:
+                        raise ValueError(
+                            '{0.name!r} is not of kind POSITIONAL_OR_KEYWORD, but:'
+                            ' {0.kind}'.format(param))
                 if param.kind == param.VAR_KEYWORD:
                     found_kws = True
                     params.extend(kwoparams)
@@ -103,8 +125,17 @@ class _PokTranslator(_util.OverrideableDataDesc):
         if not found_kws:
             params.extend(kwoparams)
         if to_use:
-            raise ValueError('leftovers ' + str(to_use)) #FIXME
-        self.__signature__ = sig.replace(parameters=params)
+            raise ValueError("Parameters not found: " + ' '.join(to_use))
+        self.__signature__ = sig.replace(
+            parameters=params,
+            sources=_signatures.copy_sources(sig.sources, {self.func:self}))
+
+    def _sigtools__autoforwards_hint(self, func):
+        ast = _util.get_ast(self.func)
+        if ast is None:
+            return None
+        sig = self.__signature__
+        return self.func, ast, sig
 
     def __call__(self, *args, **kwargs):
         intersect = self.posoarg_names.intersection(kwargs)
@@ -137,8 +168,7 @@ class _PokTranslator(_util.OverrideableDataDesc):
 
     def __repr__(self):
         return (
-            '<{0.func!r} with signature '
-            '{0.func.__name__}{0.__signature__}>'
+            '<{0.func!r} with arg translation>'
             .format(self))
 
 @_PokTranslator(kwoargs=('start',))
@@ -155,10 +185,14 @@ def kwoargs(start=None, *kwoarg_names):
         def py23_func(spam, ham, eggs='chichen'):
             return spam, ham, eggs
 
-    If ``start`` is given and is the name of a parameter, it and all
-    parameters after it are made keyword-only.
+    :param str start: If given and is the name of a parameter, it and all
+        parameters after it are made keyword-only
+    :param str kwoarg_names: Names of the parameters to convert
+
+    :raises: `ValueError` if end or one of posoarg_names isn't in the
+        decorated function's signature.
     """
-    assert all(isinstance(s, six.string_types) for s in kwoarg_names), \
+    assert all(isinstance(s, str) for s in kwoarg_names), \
         "argument names must be strings; forgot to put () after @kwoargs?"
     if start is not None:
         return partial(_kwoargs_start, start, kwoarg_names)
@@ -170,7 +204,7 @@ def kwoargs(start=None, *kwoarg_names):
 def _kwoargs_start(start, _kwoargs, func, *args, **kwargs):
     kwoarg_names = set(_kwoargs)
     found = False
-    sig = _util.signature(func).parameters.values()
+    sig = _specifiers.forged_signature(func, auto=False).parameters.values()
     for param in sig:
         if param.kind == param.POSITIONAL_OR_KEYWORD:
             if found or param.name == start:
@@ -189,8 +223,8 @@ def _kwoargs_start(start, _kwoargs, func, *args, **kwargs):
 def posoargs(end=None, *posoarg_names):
     """Marks the given parameters as positional-only.
 
-    If the resulting function is passed named arguments for any positional
-    parameter, ``TypeError`` is raised.
+    If the resulting function is passed any named arguments that references a
+    positional parameter, `TypeError` will be raised.
 
         >>> from sigtools.modifiers import posoargs
         >>> @posoargs('ham')
@@ -206,15 +240,15 @@ def posoargs(end=None, *posoarg_names):
             .format(' '.join(repr(name) for name in intersect))
         TypeError: Named arguments refer to positional-only parameters: 'ham'
 
-    If ``end`` is given and is the name of a parameter, it and all
-    parameters leading to it are made positional-only.
+    :param str end: If given and is the name of a parameter, it and all
+        parameters leading to it are made positional-only.
+    :param str posoarg_names: Names of the parameters to convert
 
-    :raises: ``ValueError`` if end or one of posoarg_names isn't in the
+    :raises: `ValueError` if end or one of posoarg_names isn't in the
         decorated function's signature.
-
     """
-    assert all(isinstance(s, six.string_types) for s in posoarg_names), \
-        "argument names must be strings; forgot to put () after @kwoargs?"
+    assert all(isinstance(s, str) for s in posoarg_names), \
+        "argument names must be strings"
     if end is not None:
         return partial(_posoargs_end, end, posoarg_names)
     if not posoarg_names:
@@ -224,7 +258,7 @@ def posoargs(end=None, *posoarg_names):
 def _posoargs_end(end, _posoargs, func, *args, **kwargs):
     posoarg_names = set(_posoargs)
     found = False
-    sig = _util.signature(func).parameters.values()
+    sig = _specifiers.forged_signature(func, auto=False).parameters.values()
     for param in sig:
         if param.kind == param.POSITIONAL_OR_KEYWORD:
             if not found:
@@ -240,22 +274,49 @@ def _posoargs_end(end, _posoargs, func, *args, **kwargs):
         func, posoargs=posoarg_names,
         get=partial(_posoargs_end, end, _posoargs))
 
-
 @kwoargs('exceptions')
-def autokwoargs(exceptions=()):
-    """Marks all arguments with default values as keyword-only."""
-    return partial(_autokwoargs, exceptions)
+def autokwoargs(func=None, exceptions=()):
+    """Marks all arguments with default values as keyword-only.
+
+    :param sequence exceptions: names of parameters not to convert
+
+    ::
+
+        >>> from sigtools.modifiers import autokwoargs
+        >>> @autokwoargs(exceptions=['c'])
+        ... def func(a, b, c=3, d=4, e=5):
+        ...     pass
+        ...
+        >>> from inspect import signature
+        >>> print(signature(func))
+        (a, b, c=3, *, d=4, e=5)
+
+    """
+    if func is not None:
+        if callable(func):
+            return _autokwoargs(exceptions, func)
+        else:
+            raise ValueError("exceptions must be passed by name")
+    else:
+        return partial(_autokwoargs, exceptions)
 
 def _autokwoargs(exceptions, func):
-    sig = _util.signature(func)
+    sig = _specifiers.forged_signature(func, auto=False)
     args = []
+    exceptions = set(exceptions)
     for param in sig.parameters.values():
         if (
                 param.kind == param.POSITIONAL_OR_KEYWORD
                 and param.default != param.empty
-                and param.name not in exceptions
             ):
-            args.append(param.name)
+            try:
+                exceptions.remove(param.name)
+            except KeyError:
+                args.append(param.name)
+    if exceptions:
+        raise ValueError(
+            "parameters referred to by 'exceptions' not present: "
+            + ' '.join(repr(name) for name in exceptions))
     return kwoargs(*args)(func)
 
 class annotate(object):
@@ -270,30 +331,35 @@ class annotate(object):
         def py23_func(spam, eggs=False):
             return spam, eggs
 
-    Raises ValueError if a parameter to be annotated does not exist
-    on the function.
+    :param _annotate__return_annotation: The annotation to attach for return
+        value
+    :param annotations: The annotations to attach for each parameter
+    :raises: `ValueError` if a parameter to be annotated does not exist
+        on the function
     """
 
     def __init__(self, __return_annotation=_util.UNSET, **annotations):
-        """
-        Initializer.
-
-        __return_annotation: The return annotation
-
-        annotations: argument=annotation pairs
-        """
         self.ret = __return_annotation
         self.annotations = annotations
         self.to_use = set(annotations)
 
-    def __call__(self, func):
-        sig = _util.signature(func)
+    def __call__(self, obj):
+        func = obj
+        poks = []
+        while isinstance(func, _PokTranslator):
+            poks.append(func)
+            func = func.func
+        sig = _specifiers.forged_signature(func, auto=False)
         parameters = []
         to_use = self.to_use.copy()
         for name, parameter in sig.parameters.items():
             if name in self.annotations:
+                annotation = self.annotations[name]
+                upgraded_annotation = _signatures.UpgradedAnnotation.preevaluated(annotation)
                 parameters.append(parameter.replace(
-                    annotation=self.annotations[name]))
+                    annotation=annotation,
+                    upgraded_annotation=upgraded_annotation,
+                ))
                 to_use.remove(name)
             else:
                 parameters.append(parameter)
@@ -306,9 +372,12 @@ class annotate(object):
             sig = sig.replace(parameters=parameters)
         else:
             sig = sig.replace(parameters=parameters,
-                              return_annotation=self.ret)
+                              return_annotation=self.ret,
+                              upgraded_return_annotation=_signatures.UpgradedAnnotation.preevaluated(self.ret))
         func.__signature__ = sig
-        return func
+        for pok in reversed(poks):
+            pok._prepare()
+        return obj
 
     def __repr__(self):
         return '{0}.annotate({1}{2})'.format(
@@ -317,3 +386,5 @@ class annotate(object):
             ', '.join('{0[0]}={0[1]!r}'.format(item)
                       for item in sorted(self.annotations.items())))
 
+
+_finalized = True
